package moonpool

  1. Overview
  2. Docs

Module Moonpool_dpoolSource

Static pool of domains.

These domains are shared between all the pools in moonpool. The rationale is that we should not have more domains than cores, so it's easier to reserve exactly that many domain slots, and run more flexible thread pools on top (each domain being shared by potentially multiple threads from multiple pools).

The pool should not contain actual domains if it's not in use, ie if no runner is presently actively using one or more of the domain slots.

The pool is initialized by the first execution request, which must be made from the main domain. A process using fork must fork before that request.

NOTE: Interface is still experimental.

  • since 0.6
Sourceval max_number_of_domains : unit -> int

Number of domains in the pool when all domains are active.

Low level interface for resouce handling

Be very cautious with this interface, or resource leaks might occur.

Sourceval run_on : int -> (unit -> unit) -> unit

run_on i f runs f() on the domain with index i. Precondition: 0 <= i < n_domains(). The first call must be made from the main domain. The thread must call decr_on with i once it's done. NOTE: if f() raises, the process will exit with error.

Sourceval decr_on : int -> unit

Signal that a thread is stopping on the domain with index i. The pool must already be initialized.

Sourceval run_on_and_wait : int -> (unit -> 'a) -> 'a

run_on_and_wait i f runs f() on the domain with index i, and blocks until the result of f() is returned back. See run_on for more details.